@@ -13,6 +13,7 @@
 #include "../libpcsxcore/psemu_plugin_defs.h"
 #include "../libpcsxcore/cdrom.h"
 #include "../libpcsxcore/cdriso.h"
+#include "../plugins/dfinput/main.h"
 #include "common/readpng.h"
 #include "maemo_common.h"
 #include <libosso.h>
@@ -545,19 +546,53 @@ inline void key_press_event(int key2,int type)
 			return;
 	}
 
-	if (type == 1) {
-		if (psxkey1 >= 0)
-			in_keystate |= 1 << psxkey1;
-		if (psxkey2 >= 0)
-			in_keystate |= 1 << psxkey2;
-	}
-	else if (type == 2) {
-		if (psxkey1 >= 0)
-			in_keystate &= ~(1 << psxkey1);
-		if (psxkey2 >= 0)
-			in_keystate &= ~(1 << psxkey2);
+	if (in_type1 == PSE_PAD_TYPE_GUNCON){
+		if (type == 1) {
+			switch (psxkey1){
+				case DKEY_CROSS:
+					in_state_gun |= SACTION_GUN_A;
+					break;		
+				case DKEY_CIRCLE:
+					in_state_gun |= SACTION_GUN_B;
+					break;		
+				case DKEY_TRIANGLE:
+					in_state_gun |= SACTION_GUN_TRIGGER2;
+					break;		
+				case DKEY_SQUARE:
+					in_state_gun |= SACTION_GUN_TRIGGER;
+					break;		
+			}
+		}else if (type == 2) {
+			switch (psxkey1){
+				case DKEY_CROSS:
+					in_state_gun &= ~SACTION_GUN_A;
+					break;		
+				case DKEY_CIRCLE:
+					in_state_gun &= ~SACTION_GUN_B;
+					break;		
+				case DKEY_TRIANGLE:
+					in_state_gun &= ~SACTION_GUN_TRIGGER2;
+					break;		
+				case DKEY_SQUARE:
+					in_state_gun &= ~SACTION_GUN_TRIGGER;
+					break;		
+			}
+		}
+	}else{
+		if (type == 1) {
+			if (psxkey1 >= 0)
+				in_keystate |= 1 << psxkey1;
+			if (psxkey2 >= 0)
+				in_keystate |= 1 << psxkey2;
+		}
+		else if (type == 2) {
+			if (psxkey1 >= 0)
+				in_keystate &= ~(1 << psxkey1);
+			if (psxkey2 >= 0)
+				in_keystate &= ~(1 << psxkey2);
 
-		emu_set_action(SACTION_NONE);
+			emu_set_action(SACTION_NONE);
+		}
 	}
 }
 
@@ -610,6 +645,22 @@ int maemo_init(int *argc, char ***argv)
 		printf("Game Gripper detected\n");
 	}
 
+	switch (in_type1){
+		case PSE_PAD_TYPE_GUNCON:
+			memset(cornerActions, 0, sizeof(cornerActions));
+			printf("Controller set to GUNCON (SLPH-00034)\n");
+			break;
+		case PSE_PAD_TYPE_STANDARD:
+			printf("Controller set to standard (SCPH-1080)\n");
+			break;
+		case PSE_PAD_TYPE_ANALOGPAD:
+			printf("Controller set to analog (SCPH-1150)\n");
+			break;	
+	}
+
+	if (in_enable_vibration)
+		printf("Vibration enabled\n");
+
 	if (!(g_maemo_opts&8)){
 		gtk_init (argc, argv);
 
@@ -790,8 +841,20 @@ void plat_step_volume(int is_up)
 {
 }
 
+//#include <sys/time.h>
+//static double last_vibrate = 0;
 void plat_trigger_vibrate(int is_strong)
 {
+	const int vDuration = 10;
+	/*struct timeval tim;
+	gettimeofday(&tim, NULL);
+	double t1 = tim.tv_sec + (tim.tv_usec / 1000000.0);
+	if (last_vibrate && t1 - last_vibrate < vDuration / 1000.0){
+		//printf("Don't vibrate (%f)\n", t1 - last_vibrate);
+		return;
+	}
+	last_vibrate = t1;*/
+
 	DBusConnection* system_bus = (DBusConnection*)osso_get_sys_dbus_connection(osso);
 	DBusMessageIter args;
 	DBusMessage*msg = dbus_message_new_method_call("com.nokia.mce",
@@ -800,8 +863,8 @@ void plat_trigger_vibrate(int is_strong)
 												   "req_start_manual_vibration");
 	if (msg) {
 		dbus_message_iter_init_append(msg, &args);
-		int speed = is_strong ? 200 : 100;
-		int duration = 10;
+		int speed = is_strong ? 200 : 150;
+		int duration = vDuration;
 		if (dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32, &speed)) {
 			if (dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32, &duration)) {
 				dbus_connection_send(system_bus, msg, NULL);
